#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Functions for handling Nala History."""
from __future__ import annotations

import json
import sys
from getpass import getuser
from json.decoder import JSONDecodeError
from os import environ, getuid
from pwd import getpwnam
from typing import Dict, Generator, Iterable, List, Tuple, Union, cast

import typer
from apt.package import Package

from nala import _, color
from nala.cache import Cache
from nala.constants import ERROR_PREFIX, NALA_HISTORY, WARNING_PREFIX
from nala.options import (
	ASSUME_YES,
	AUTO_REMOVE,
	DEBUG,
	DOWNLOAD_ONLY,
	FIX_BROKEN,
	MAN_HELP,
	OPTION,
	PURGE,
	RAW_DPKG,
	RECOMMENDS,
	REMOVE_ESSENTIAL,
	SIMPLE,
	SUGGESTS,
	UPDATE,
	VERBOSE,
	arguments,
	history_typer,
)
from nala.rich import ELLIPSIS, OVERFLOW, Column, Table
from nala.summary import print_update_summary
from nala.utils import (
	DelayedKeyboardInterrupt,
	NalaPackage,
	PackageHandler,
	ask,
	command_help,
	dprint,
	eprint,
	get_date,
	term,
)

USER: str = environ.get("DOAS_USER", "")
UID: int = 0
if USER:
	UID = getpwnam(USER).pw_uid
else:
	USER = environ.get("SUDO_USER", getuser())
	UID = int(environ.get("SUDO_UID", getuid()))

HistoryFile = Dict[str, Dict[str, Union[str, bool, List[str], List[List[str]]]]]
HistoryEntry = Dict[str, Union[str, bool, List[str], List[List[str]]]]

NOT_SUPPORTED = _(
	"{error} '{command}' for operations other than install or remove are not currently supported"
)

HISTORY_HELP = _(
	"Show transaction history.\n\n"
	"Running `nala history` with no subcommands prints an overview of all transactions."
)


def load_history_file() -> HistoryFile:
	"""Load Nala history."""
	try:
		return cast(
			HistoryFile,
			json.loads(NALA_HISTORY.read_text(encoding="utf-8")),
		)
	except JSONDecodeError:
		sys.exit(
			_(
				"{error} History file seems corrupt. You should try removing {file}"
			).format(error=ERROR_PREFIX, file=NALA_HISTORY)
		)
	except PermissionError as error:
		sys.exit(f"{error}")


def write_history_file(data: HistoryFile) -> None:
	"""Write history to file."""
	with DelayedKeyboardInterrupt():
		with open(NALA_HISTORY, "w", encoding="utf-8") as file:
			file.write(json.dumps(data, separators=(",", ":")))


def get_history(hist_id: str) -> HistoryEntry:
	"""Get the history from file."""
	dprint(f"Getting History Entry: {hist_id}")
	if not NALA_HISTORY.exists():
		sys.exit(_("{error} No history exists.").format(error=ERROR_PREFIX))
	if transaction := load_history_file().get(hist_id):
		return transaction
	sys.exit(
		_("{error} Transaction {num} doesn't exist.").format(
			error=ERROR_PREFIX, num=hist_id
		)
	)


def nala_installed(value: bool) -> None:
	"""Print packages that are explicitly installed by Nala."""
	if not value:
		return
	user_installed = get_list(get_history("Nala"), "User-Installed")
	for pkg in user_installed:
		print(pkg)
	sys.exit()


def get_nala_packages(hist_entry: HistoryEntry, key: str) -> list[NalaPackage]:
	"""Type cast history package is list of lists."""
	nala_pkgs = []
	for pkg_list in get_packages(hist_entry, key):
		dprint(f"{key} List: {pkg_list}")
		if len(pkg_list) == 4:
			try:
				name, new_version, size, old_version = pkg_list
				nala_pkgs.append(NalaPackage(name, new_version, int(size), old_version))
			except ValueError:
				name, old_version, new_version, size = pkg_list
				nala_pkgs.append(NalaPackage(name, new_version, int(size), old_version))
			continue
		name, new_version, size = pkg_list
		nala_pkgs.append(NalaPackage(name, new_version, int(size)))
	return nala_pkgs


def get_packages(hist_entry: HistoryEntry, key: str) -> list[list[str]]:
	"""Type cast history packages is list of strings."""
	return cast(List[List[str]], hist_entry.get(key, [[]]))


def get_list(hist_entry: HistoryEntry, key: str) -> list[str]:
	"""Type cast history command is list of strings."""
	item = cast(List[str], hist_entry.get(key, []))
	dprint(f"Getting List {key} {item}")
	return item


def get_bool(hist_entry: HistoryEntry, key: str) -> bool:
	"""Type cast history entry is bool."""
	return cast(bool, hist_entry.get(key, False))


def get_str(hist_entry: HistoryEntry, key: str) -> str:
	"""Type cast history entry is str."""
	return cast(str, hist_entry.get(key, ""))


def pop_nala(hist_file: HistoryFile) -> HistoryEntry:
	"""Pop the Nala field from the history."""
	return hist_file.pop("Nala", {})


def get_last(hist_file: HistoryFile) -> HistoryEntry:
	"""Return the last entry in the history file."""
	internal = hist_file.copy()
	# Pop the Nala system information as to not count it.
	pop_nala(internal)
	# Return the last entry in the history file.
	return hist_file.get(f"{len(internal)}", {})


def set_user_installed(
	cache: Cache, user_explicit: list[Package], user_installed: set[str]
) -> None:
	"""Set the User-Installed field."""
	if user_explicit:
		for pkg in user_explicit:
			if pkg.marked_install:
				user_installed.add(pkg.name)
			if pkg.marked_delete:
				user_installed.discard(pkg.name)

	user_installed = {
		pkg_name
		for pkg_name in user_installed
		if pkg_name in cache
		and (cache[pkg_name].installed or cache[pkg_name].marked_install)
	}


def write_history(cache: Cache, handler: PackageHandler, operation: str) -> None:
	"""Prepare history for writing."""
	history_dict = load_history_file() if NALA_HISTORY.exists() else {}
	nala_dict = pop_nala(history_dict)
	user_installed = set(get_list(nala_dict, "User-Installed"))

	# Make sure the numbers of the history entries are concurrent.
	# Otherwise a history entry will forever be overwritten each transaction.
	history_dict = {
		f"{num + 1}": value for num, value in enumerate(history_dict.values())
	}

	hist_id = f"{len(history_dict) + 1 if history_dict else 1}"
	altered = (
		len(handler.delete_pkgs + handler.delete_config)
		+ len(handler.autoremove_pkgs + handler.autoremove_config)
		+ len(handler.install_pkgs)
		+ len(handler.upgrade_pkgs)
		+ len(handler.downgrade_pkgs)
		+ len(handler.reinstall_pkgs)
		+ len(handler.configure_pkgs)
		+ len(handler.local_debs)
	)

	set_user_installed(cache, handler.user_explicit, user_installed)

	transaction: HistoryEntry = {
		"Date": get_date(),
		"Requested-By": f"{USER} ({UID})",
		"Command": sys.argv[1:],
		"Altered": f"{altered}",
		"Purged": arguments.is_purge(),
		"Operation": operation,
		"Explicit": [pkg.name for pkg in handler.user_explicit],
		"Removed": [
			[pkg.name, pkg.version, f"{pkg.size}"]
			for pkg in handler.delete_pkgs + handler.delete_config
		],
		"Auto-Removed": [
			[pkg.name, pkg.version, f"{pkg.size}"]
			for pkg in handler.autoremove_pkgs + handler.autoremove_config
		],
		"Installed": [
			[pkg.name, pkg.version, f"{pkg.size}"] for pkg in handler.install_pkgs
		],
		"Reinstalled": [
			[pkg.name, pkg.version, f"{pkg.size}"] for pkg in handler.reinstall_pkgs
		],
		"Upgraded": [
			[pkg.name, pkg.version, f"{pkg.size}", f"{pkg.old_version}"]
			for pkg in handler.upgrade_pkgs
		],
		"Downgraded": [
			[pkg.name, pkg.version, f"{pkg.size}", f"{pkg.old_version}"]
			for pkg in handler.downgrade_pkgs
		],
	}

	history_dict[hist_id] = transaction
	nala_dict["History-Version"] = "1"
	nala_dict["User-Installed"] = list(user_installed)
	history_dict["Nala"] = nala_dict
	write_history_file(history_dict)


def hist_id_completion() -> Generator[Tuple[str, str], None, None]:
	"""Complete history ID arguments."""
	if not NALA_HISTORY.exists():
		return
	history_file = load_history_file()
	pop_nala(history_file)
	for key, entry in history_file.items():
		if (command := get_list(entry, "Command"))[0] in ("update", "upgrade"):
			command.extend(pkg.name for pkg in get_nala_packages(entry, "Upgraded"))
		yield key, " ".join(command)


HIST_ID = typer.Argument(
	..., metavar="ID", help=_("Transaction number"), autocompletion=hist_id_completion
)


@history_typer.callback(invoke_without_command=True, help=HISTORY_HELP)
# pylint: disable=unused-argument
def history_summary(
	ctx: typer.Context,
	installed: bool = typer.Option(
		False,
		"--installed",
		callback=nala_installed,
		help=_("Show packages that were explicitly installed with Nala"),
	),
	debug: bool = DEBUG,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
) -> None:
	"""Show transaction history.

	Running `nala history` with no subcommands prints an overview of all translations.
	"""
	# Stop this function from running if we do `nala history info`
	if ctx.invoked_subcommand:
		return

	if not NALA_HISTORY.exists():
		sys.exit(_("{error} No history exists.").format(error=ERROR_PREFIX))
	history_file = load_history_file()
	pop_nala(history_file)

	names: list[Iterable[str]] = []

	for key, entry in history_file.items():
		dprint(f"History ID {key}")
		if (command := get_list(entry, "Command"))[0] in ("update", "upgrade"):
			command.extend(pkg.name for pkg in get_nala_packages(entry, "Upgraded"))

		names.append(
			(
				key,
				" ".join(command),
				get_str(entry, "Date"),
				get_str(entry, "Altered"),
				get_str(entry, "Requested-By"),
			)
		)
	if not names:
		sys.exit(_("{error} No history exists.").format(error=ERROR_PREFIX))

	max_width = term.columns - 69
	history_table = Table(
		Column("ID"),
		Column("Command", no_wrap=True, max_width=max_width, overflow=OVERFLOW),
		Column("Date and Time", no_wrap=True),
		Column("Altered", justify="right"),
		Column("Requested-By"),
		padding=(0, 2),
		box=None,
	)

	for item in names:
		history_table.add_row(*item)
	term.console.print(history_table)


@history_typer.command("info", help=_("Show information about a specific transaction."))
@history_typer.command("show", hidden=True)
# pylint: disable=unused-argument
def history_info(
	ctx: typer.Context,
	hist_id: str = HIST_ID,
	simple: bool = SIMPLE,
	debug: bool = DEBUG,
	verbose: bool = VERBOSE,
) -> None:
	"""Show information about a specific transaction."""
	arguments.history = ctx.command.name
	command_help("show", "history info", None)

	hist_entry = (
		get_last(load_history_file())
		if hist_id.lower() == "last"
		else get_history(hist_id)
	)

	dprint(f"History Entry: {hist_entry}")
	arguments.purge = get_bool(hist_entry, "Purged")
	nala_pkgs = PackageHandler()
	nala_pkgs.autoremove_pkgs = get_nala_packages(hist_entry, "Auto-Removed")
	nala_pkgs.delete_pkgs = get_nala_packages(hist_entry, "Removed")
	nala_pkgs.install_pkgs = get_nala_packages(hist_entry, "Installed")
	nala_pkgs.reinstall_pkgs = get_nala_packages(hist_entry, "Reinstalled")
	nala_pkgs.upgrade_pkgs = get_nala_packages(hist_entry, "Upgraded")
	nala_pkgs.downgrade_pkgs = get_nala_packages(hist_entry, "Downgraded")

	print_update_summary(nala_pkgs)


def history_sudo(
	redo: bool = False,
	clear: bool = False,
) -> None:
	"""Check if we need sudo."""
	if not term.is_su():
		if clear:
			sys.exit(" ".join([ERROR_PREFIX, _("Nala needs root to clear history")]))
		if redo:
			sys.exit(" ".join([ERROR_PREFIX, _("Nala needs root to redo history")]))
		sys.exit(" ".join([ERROR_PREFIX, _("Nala needs root to undo history")]))


def unlink_history(value: bool) -> None:
	"""Remove the history file."""
	if not value:
		return
	history_sudo(clear=True)
	dprint("History clear all")

	history_dict = {"Nala": get_history("Nala")}
	NALA_HISTORY.unlink(missing_ok=True)
	write_history_file(history_dict)

	print(_("History has been cleared"))
	sys.exit()


@history_typer.command("clear", help=_("Clear a transaction or the entire history."))
# pylint: disable=unused-argument
def history_clear(
	_hist_id: int = HIST_ID,
	_all: bool = typer.Option(  # pylint: disable=unused-argument
		False, "--all", callback=unlink_history, help=_("Clear the entire history.")
	),
	debug: bool = DEBUG,
	verbose: bool = VERBOSE,
) -> None:
	"""Clear a transaction or the entire history."""
	hist_id = f"{_hist_id}"
	dprint(f"History clear {hist_id}")
	if not NALA_HISTORY.exists():
		eprint(_("No history exists to clear") + ELLIPSIS)
		return

	if hist_id not in (history_file := load_history_file()).keys():
		sys.exit(
			_("{error} ID: {hist_id} does not exist in the history").format(
				error=ERROR_PREFIX, hist_id=color(hist_id, "YELLOW")
			)
		)
	history_edit: HistoryFile = {}
	nala_dict = pop_nala(history_file)
	num = 0
	# Using sum increments to relabled the IDs so when you remove just one
	# There isn't a gap in ID numbers and it looks concurrent.
	for key, value in history_file.items():
		if key != hist_id:
			num += 1
			history_edit[f"{num}"] = value
	history_edit["Nala"] = nala_dict
	write_history_file(history_edit)
	print(_("History has been altered") + ELLIPSIS)


@history_typer.command("undo", help=_("Undo a transaction."))
@history_typer.command("redo", help=_("Redo a transaction."))
# pylint: disable=unused-argument,too-many-arguments,too-many-locals
def history_undo(
	ctx: typer.Context,
	hist_id: str = HIST_ID,
	purge: bool = PURGE,
	debug: bool = DEBUG,
	raw_dpkg: bool = RAW_DPKG,
	download_only: bool = DOWNLOAD_ONLY,
	remove_essential: bool = REMOVE_ESSENTIAL,
	update: bool = UPDATE,
	auto_remove: bool = AUTO_REMOVE,
	install_recommends: bool = RECOMMENDS,
	install_suggests: bool = SUGGESTS,
	fix_broken: bool = FIX_BROKEN,
	assume_yes: bool = ASSUME_YES,
	simple: bool = SIMPLE,
	dpkg_option: List[str] = OPTION,
	verbose: bool = VERBOSE,
) -> None:
	"""History undo/redo commands."""
	from nala.nala import (  # pylint: disable=cyclic-import, import-outside-toplevel
		_install,
		_remove,
	)

	arguments.history = ctx.command.name
	arguments.history_id = hist_id
	redo = ctx.command.name == "redo"
	history_sudo(redo=redo)

	dprint(f"History: {ctx.command.name} {hist_id}")

	transaction = (
		get_last(load_history_file())
		if hist_id.lower() == "last"
		else get_history(hist_id)
	)

	dprint(f"Transaction: {transaction}")
	if not arguments.purge:
		if purge := get_bool(transaction, "Purged"):
			eprint(
				_("{warning} This history entry was a purge.").format(
					warning=WARNING_PREFIX
				)
			)
			if ask(_("Do you want to continue with purge enabled?")):
				arguments.purge = purge

	explicit = get_list(transaction, "Explicit")
	operation = get_str(transaction, "Operation")
	command = get_list(transaction, "Command")[0]

	if operation == "remove" or command in (
		"remove",
		"purge",
		"autoremove",
		"autopurge",
	):
		pkgs = [pkg[0] for pkg in get_packages(transaction, "Removed")]
		pkgs.extend([pkg[0] for pkg in get_packages(transaction, "Auto-Removed")])

		if redo:
			_remove(pkgs)
			return
		_install(explicit or pkgs, ctx)
		return

	if operation == "install" or command == "install":
		pkgs = [pkg[0] for pkg in get_packages(transaction, "Installed")]
		if redo:
			_install(explicit or pkgs, ctx)
			return
		_remove(pkgs)
		return
	sys.exit(
		NOT_SUPPORTED.format(
			error=ERROR_PREFIX, command=f"{arguments.command} {ctx.command.name}"
		)
	)



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Module for printing the transaction summary."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Generator, Iterable

from nala import _, color, console
from nala.cache import Cache
from nala.options import arguments
from nala.rich import HORIZONTALS, OVERFLOW, Column, Group, Table, Text, Tree, from_ansi
from nala.utils import NalaPackage, PackageHandler, dprint, term, unit_str

# NOTE: The following are the headers for the transaction summary.
# NOTE: Package:        Version:     Size:
# NOTE: ansible-core    2.12.4-1    1.2 MB
PACKAGE, VERSION, SIZE, OLD_VERSION, NEW_VERSION, EITHER = _(
	"Package/Version/Size/Old Version/New Version/Either"
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Auto-Purge 20 Packages" ]
_AUTO_PURGE, _AUTO_PURGING, _AUTO_PURGED = _(
	"Auto-Purge/Auto-Purging/Auto-Purged",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Auto-Remove 20 Packages" ]
_AUTO_REMOVE, _AUTO_REMOVING, _AUTO_REMOVED = _(
	"Auto-Remove/Auto-Removing/Auto-Removed",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Remove 20 Packages" ]
_REMOVE, _REMOVING, _REMOVED = _(
	"Remove/Removing/Removed",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Purge 20 Packages" ]
_PURGE, _PURGING, _PURGED = _(
	"Purge/Purging/Purged",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Install 20 Packages" ]
_INSTALL, _INSTALLING, _INSTALLED = _(
	"Install/Installing/Installed",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Reinstall 20 Packages" ]
_REINSTALL, _REINSTALLING, _REINSTALLED = _(
	"Reinstall/Reinstalling/Reinstalled",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Upgrade 20 Packages" ]
_UPGRADE, _UPGRADING, _UPGRADED = _(
	"Upgrade/Upgrading/Upgraded",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Downgrade 20 Packages" ]
_DOWNGRADE, _DOWNGRADING, _DOWNGRADED = _(
	"Downgrade/Downgrading/Downgraded",
).split("/")
# NOTE: Verb Tenses are [ "Present/Present Participle/Past" ]
# NOTE: This ends up looking like [ "Configure 20 Packages" ]
_CONFIGURE, _CONFIGURING, _CONFIGURED = _(
	"Configure/Configuring/Configured",
).split("/")

SUMMARY_LAYOUT = ("left_adjust", "right_adjust", "left_adjust")
UPGRADE_LAYOUT = ("pkg_blue", "old_version", "new_version", "pkg_size")
DOWNGRADE_LAYOUT = ("pkg_yellow", "old_version", "new_version", "pkg_size")
HELD_LAYOUT = ("pkg_yellow", "old_version", "new_version", "pkg_size")
DEFAULT_LAYOUT = ("pkg_green", "version", "pkg_size")
EXTRA_LAYOUT = ("pkg_magenta", "version", "pkg_size")
REMOVE_LAYOUT = ("pkg_red", "version", "pkg_size")

COLUMN_MAP: dict[str, dict[str, str | int]] = {
	"left_adjust": {"overflow": OVERFLOW},
	"right_adjust": {"justify": "right", "overflow": OVERFLOW},
	"pkg_green": {
		"header": f"{PACKAGE}:",
		"style": "bold green",
		"color": "GREEN",
		"overflow": OVERFLOW,
		"ratio": 2,
	},
	"pkg_red": {
		"header": f"{PACKAGE}:",
		"style": "bold red",
		"color": "RED",
		"overflow": OVERFLOW,
		"ratio": 2,
	},
	"pkg_blue": {
		"header": f"{PACKAGE}:",
		"style": "bold blue",
		"color": "BLUE",
		"overflow": OVERFLOW,
		"ratio": 3,
	},
	"pkg_yellow": {
		"header": f"{PACKAGE}:",
		"style": "bold orange_red1",
		"color": "YELLOW",
		"overflow": OVERFLOW,
		"ratio": 3,
	},
	"pkg_magenta": {
		"header": f"{PACKAGE}:",
		"style": "bold magenta",
		"color": "MAGENTA",
		"overflow": OVERFLOW,
		"ratio": 2,
	},
	"pkg_size": {
		"header": f"{SIZE}:",
		"justify": "right",
		"overflow": OVERFLOW,
		"ratio": 2,
	},
	"version": {"header": f"{VERSION}:", "overflow": "fold", "ratio": 2},
	"old_version": {"header": f"{OLD_VERSION}:", "overflow": "fold", "ratio": 2},
	"new_version": {"header": f"{NEW_VERSION}:", "overflow": "fold", "ratio": 2},
}
ROW_MAP: dict[str, str] = {
	"pkg_green": "name",
	"pkg_red": "name",
	"pkg_blue": "name",
	"pkg_yellow": "name",
	"pkg_magenta": "name",
	"pkg_size": "unit_size",
	"version": "version",
	"old_version": "old_version",
	"new_version": "version",
}


def get_columns(column_keys: Iterable[str]) -> Generator[Column, None, None]:
	"""Get the columns from our column map."""
	for key in column_keys:
		# Have to pop the color key to work with Column
		if "color" in (kwargs := COLUMN_MAP[key]):
			kwargs.pop("color")
		yield Column(**kwargs)  # type: ignore[arg-type]


def get_rows(pkg: NalaPackage, layout: Iterable[str]) -> Generator[Text, None, None]:
	"""Get the rows from our row map."""
	for key in layout:
		if key == "new_version":
			yield from_ansi(version_diff(pkg))
			continue
		if key == "color":
			continue
		yield from_ansi(getattr(pkg, ROW_MAP[key]))


def version_diff(pkg: NalaPackage) -> str:
	"""Return a colored diff of the new version."""
	# If there is no old version, or the versions are the same
	# then don't color anything
	if not pkg.old_version or pkg.old_version == pkg.version:
		return pkg.version

	# Check for just revision change first.
	old_ver = pkg.old_version.rsplit("-", 1)
	new_ver = pkg.version.rsplit("-", 1)

	# If there isn't a revision these shouldn't ever match
	# If they do match then only the revision has changed
	if old_ver[0] == new_ver[0]:
		return f"{new_ver[0]}-{color(new_ver[1], 'YELLOW')}"

	# Split both version strings. Example "1.12.3" = ["1", "12", "3"]
	old_ver = pkg.old_version.split(".", 2)
	new_ver = pkg.version.split(".", 2)

	start_color = 0
	for i, section in enumerate(old_ver):
		# Guard clause, no index error below
		if i > len(new_ver) - 1:
			break

		if section != new_ver[i]:
			start_color = i
			break

	# Rebuild the version string with color for diff sections
	colored_ver = []
	for i, section in enumerate(new_ver):
		if i >= start_color:
			colored_ver.append(color(section, "YELLOW"))
			continue
		colored_ver.append(section)
	return ".".join(colored_ver)


@dataclass
class PackageHeaders:
	"""Tuple for package headers."""

	layout: tuple[str, ...]
	title: str
	summary: str = ""


@dataclass
class Headers:  # pylint: disable=too-many-instance-attributes
	"""Tuple for headers."""

	deleting: PackageHeaders
	auto_removing: PackageHeaders
	installing: PackageHeaders
	reinstalling: PackageHeaders
	upgrading: PackageHeaders
	downgrading: PackageHeaders
	configuring: PackageHeaders
	recommending: PackageHeaders
	suggesting: PackageHeaders
	held_pkgs: PackageHeaders
	# Packages that can be auto removed, but won't
	not_needed: PackageHeaders | None = None


def auto_remove_header(history: bool) -> tuple[str, str]:
	"""Get the auto remove header."""
	if history and arguments.is_purge():
		# NOTE: `/` will convert to a space for example in Irish it may be:
		# NOTE: "{package_total}/bpacáistí/a uasghrádú" -> "20 bpacáistí a uasghrádú"
		return _AUTO_PURGED, _AUTO_PURGED
	if history:
		return _AUTO_REMOVED, _AUTO_REMOVED
	if arguments.is_purge():
		return _AUTO_PURGING, _AUTO_PURGE
	return _AUTO_REMOVING, _AUTO_REMOVE


def remove_header(history: bool) -> tuple[str, str]:
	"""Get the remove header."""
	if history and arguments.is_purge():
		return _PURGED, _PURGED
	if history:
		return _REMOVED, _REMOVED
	return (_PURGING, _PURGE) if arguments.is_purge() else (_REMOVING, _REMOVE)


def get_headers() -> Headers:
	"""Get the headers for the summary table."""
	return Headers(
		PackageHeaders(REMOVE_LAYOUT, *remove_header(history=False)),
		PackageHeaders(REMOVE_LAYOUT, *auto_remove_header(history=False)),
		PackageHeaders(DEFAULT_LAYOUT, _INSTALLING, _INSTALL),
		PackageHeaders(DEFAULT_LAYOUT, _REINSTALLING, _REINSTALL),
		PackageHeaders(UPGRADE_LAYOUT, _UPGRADING, _UPGRADE),
		PackageHeaders(DOWNGRADE_LAYOUT, _DOWNGRADING, _DOWNGRADE),
		PackageHeaders(EXTRA_LAYOUT, _CONFIGURING, _CONFIGURE),
		PackageHeaders(EXTRA_LAYOUT, _("Recommended, Will Not Be Installed")),
		PackageHeaders(EXTRA_LAYOUT, _("Suggested, Will Not Be Installed")),
		PackageHeaders(
			DOWNGRADE_LAYOUT, _("Kept Back, Will Not Be Upgraded"), _("Kept Back")
		),
		PackageHeaders(REMOVE_LAYOUT, _("Auto-Removable, Will Not Be Removed")),
	)


def get_history_headers() -> Headers:
	"""Get the headers for the summary table."""
	return Headers(
		PackageHeaders(REMOVE_LAYOUT, *remove_header(history=True)),
		PackageHeaders(REMOVE_LAYOUT, *auto_remove_header(history=True)),
		PackageHeaders(DEFAULT_LAYOUT, _INSTALLED, _INSTALLED),
		PackageHeaders(DEFAULT_LAYOUT, _REINSTALLED, _REINSTALLED),
		PackageHeaders(UPGRADE_LAYOUT, _UPGRADED, _UPGRADED),
		PackageHeaders(DOWNGRADE_LAYOUT, _DOWNGRADED, _DOWNGRADED),
		PackageHeaders(EXTRA_LAYOUT, _CONFIGURED, _CONFIGURED),
		PackageHeaders(EXTRA_LAYOUT, _("Recommended, Will Not Be Installed")),
		PackageHeaders(
			DOWNGRADE_LAYOUT, _("Kept Back, Will Not Be Upgraded"), _("Kept Back")
		),
		PackageHeaders(EXTRA_LAYOUT, _("Suggested, Will Not Be Installed")),
	)


def gen_printers(
	nala_pkgs: PackageHandler, headers: Headers
) -> Generator[tuple[list[NalaPackage], PackageHeaders], None, None]:
	"""Generate the printers."""
	yield from (  # type: ignore[misc]
		(nala_pkgs.not_needed, headers.not_needed),
		(nala_pkgs.install_pkgs, headers.installing),
		(nala_pkgs.reinstall_pkgs, headers.reinstalling),
		(nala_pkgs.upgrade_pkgs, headers.upgrading),
		(nala_pkgs.downgrade_pkgs, headers.downgrading),
		(nala_pkgs.configure_pkgs, headers.configuring),
		(nala_pkgs.recommend_pkgs, headers.recommending),
		(nala_pkgs.suggest_pkgs, headers.suggesting),
		(nala_pkgs.held_pkgs, headers.held_pkgs),
		(
			nala_pkgs.autoremove_pkgs + nala_pkgs.autoremove_config,
			headers.auto_removing,
		),
		(nala_pkgs.delete_pkgs + nala_pkgs.delete_config, headers.deleting),
	)


def gen_package_table(
	nala_packages: list[NalaPackage] | list[NalaPackage | list[NalaPackage]],
	pkg_headers: PackageHeaders,
) -> Table:
	"""Print package transactions in a pretty format."""
	package_table = Table(
		*get_columns(pkg_headers.layout), padding=(0, 1), box=None, expand=True
	)

	# Add our packages
	or_deps: list[list[NalaPackage]] = []
	for pkg in nala_packages:
		if isinstance(pkg, list):
			or_deps.append(pkg)
			continue
		if pkg.old_version:
			package_table.add_row(*get_rows(pkg, pkg_headers.layout))
			continue
		package_table.add_row(*get_rows(pkg, pkg_headers.layout))

	# Add any or_deps
	for pkg in or_deps:
		package_table.add_row(*summary_or_depends(pkg))

	return package_table


def summary_or_depends(pkg: list[NalaPackage]) -> tuple[Tree, Group, Group]:
	"""Format Recommend and Suggests or dependencies."""
	pkg_tree = Tree(f"[default]{EITHER}[/default]", guide_style="default")
	_children = [pkg_tree.add(npkg.name) for npkg in pkg]
	return (
		pkg_tree,
		Group("", *(npkg.version for npkg in pkg)),
		Group("", *(npkg.unit_size for npkg in pkg)),
	)


def print_update_summary(nala_pkgs: PackageHandler, cache: Cache | None = None) -> None:
	"""Print our transaction summary."""
	dprint("Printing Update Summary")
	if arguments.simple_summary:
		print_short_summary(nala_pkgs, cache)
		return

	headers = get_headers() if cache else get_history_headers()

	main_table = Table.grid(expand=True)
	summary_header = Table(_("Summary"), padding=0, box=HORIZONTALS, expand=True)
	summary_table = Table.grid(*get_columns(SUMMARY_LAYOUT), padding=(0, 1))

	for pkg_set, header in gen_printers(nala_pkgs, headers):
		if not pkg_set:
			continue
		package_table = Table(header.title, padding=0, box=HORIZONTALS, expand=True)
		package_table.add_row(gen_package_table(pkg_set, header))
		main_table.add_row(package_table)
		# We don't need empty rows from these in the summary
		if nala_pkgs.no_summary(pkg_set):
			continue
		# NOTE: This ends up looking like [ "Configure 20 Packages" ]
		summary_table.add_row(header.summary, f"{len(pkg_set)}", _("Packages"))

	summary_header.add_row(summary_table)
	main_table.add_row(summary_header)
	console.print(main_table)

	if cache:
		footer_table = Table(
			*get_columns(SUMMARY_LAYOUT), box=None, show_footer=True, show_header=False
		)
		if (download := cache.required_download) > 0:
			footer_table.add_row(_("Total download size"), unit_str(download))
		if (space := cache.required_space) < 0:
			footer_table.add_row(_("Disk space to free"), unit_str(-space))
		if space > 0:
			footer_table.add_row(_("Disk space required"), unit_str(space))
		console.print(footer_table)

	if cache and arguments.download_only:
		print(_("Nala will only download the packages"))


def append_or_print(string: str, pkg_name: str) -> bool:
	"""Print the string and return False, or return True."""
	string_size = len(from_ansi(string))
	pkg_name_size = len(from_ansi(pkg_name))

	# Check the string size plus a little buffer
	# against the available terminal width
	if string_size + pkg_name_size + 3 >= term.columns:
		dprint(
			f"Line Buffer Full: {string_size + pkg_name_size + 3},"
			f" Terminal Width: {term.columns}; Printing Buffer"
		)
		return False
	return True


def format_pkgs(
	pkg_set: list[NalaPackage] | list[NalaPackage | list[NalaPackage]],
	pkg_color: str,
) -> None:
	"""Format the packages in a simple way."""
	pkg_string = ""
	final_i = len(pkg_set) - 1
	for i, pkg in enumerate(pkg_set):
		strip_comma = final_i == i

		# We don't want to add the space at the start of the string
		if pkg_string:
			pkg_string += " "

		if isinstance(pkg, list):
			next_string = f"[ {' | '.join(pkg.name for pkg in pkg)} ],"
			if append_or_print(pkg_string, next_string):
				pkg_string += next_string
			if not strip_comma:
				continue

		if isinstance(pkg, NalaPackage) and append_or_print(pkg_string, pkg.name):
			pkg_string += f"{color(pkg.name, pkg_color)},"
			if not strip_comma:
				continue

		dprint(
			f"Current Index: {i}, Final Index: {final_i}, Strip Comma: {strip_comma}"
		)
		print(f"  {pkg_string.strip(', ') if strip_comma else pkg_string}")
		pkg_string = ""


def print_short_summary(nala_pkgs: PackageHandler, cache: Cache | None = None) -> None:
	"""Print our transaction summary."""
	dprint("Printing Update Summary")
	headers = get_headers() if cache else get_history_headers()

	summary_table = []
	for pkg_set, header in gen_printers(nala_pkgs, headers):
		if not pkg_set or not header:
			continue

		dprint(f"{header.title}: {pkg_set}")
		# Not sure on this formatting yet
		print(color(f"{header.title}:"))
		# The color should be the first index of our layout.
		# This might be "pkg_magenta". And then get the nala 'color' key
		# Ensure it is a string with a quick cast. There could be ints
		pkg_color = f"{COLUMN_MAP[header.layout[0]]['color']}"
		format_pkgs(pkg_set, pkg_color)

		# No summary is needed for this one.
		if nala_pkgs.no_summary(pkg_set):
			continue
		# NOTE: This ends up looking like [ "Configure 20 Packages" ]
		summary_table.append(f"{color(header.summary, pkg_color)} {len(pkg_set)}")

	# Print the summary
	if summary_table:
		summary = color(_("Summary"))
		print(f"{summary}:\n  {', '.join(summary_table)}")

	if cache:
		cache_string = "  "
		if (download := cache.required_download) > 0:
			cache_string += color(_("Total download size")) + f" {unit_str(download)}, "
		if (space := cache.required_space) < 0:
			cache_string += color(_("Disk space to free")) + f" {unit_str(-space)}, "
		if space > 0:
			cache_string += color(_("Disk space required")) + f" {unit_str(space)}, "
		print(cache_string.rstrip(", "))

	if cache and arguments.download_only:
		print(_("Nala will only download the packages"))



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Functions for the Nala Search command."""
from __future__ import annotations

from typing import Iterable, Pattern

from apt import Cache, Package, Version

from nala import COLOR_CODES, _, color
from nala.options import arguments
from nala.rich import ascii_replace, is_utf8
from nala.utils import get_version, pkg_candidate, pkg_installed

TOP_LINE = "├──" if is_utf8 else "+--"
BOT_LINE = "└──" if is_utf8 else "`--"
LINE = "│   " if is_utf8 else "|   "


def skip_pkg(
	cache: Cache, pkg: Package, nala_installed: bool, user_installed: list[str]
) -> bool:
	"""Check if the package should be skipped based on user switches."""
	if nala_installed and pkg.shortname not in user_installed:
		return True
	if arguments.installed and not pkg.installed:
		return True
	if arguments.upgradable and not pkg.is_upgradable:
		return True
	return bool(arguments.virtual and not cache.is_virtual_package(pkg.name))


def search_name(pkg: Package, patterns: list[Pattern[str]]) -> bool:
	"""Search the package name and description."""
	searches = [pkg.shortname]
	if not arguments.names:
		records = pkg._pcache._records
		records.lookup(pkg._pkg.version_list[0].file_list[0])
		searches.extend([records.long_desc, records.source_pkg])

	for pattern in patterns:
		for string in searches:
			if pattern.search(string):
				return True
	return False


def iter_search(found: Iterable[Package]) -> bool:
	"""Iterate the search results."""
	pkg_list_check: list[Package] = []
	for pkg in found:
		version = get_version(pkg, inst_first=True)
		# Version being a tuple here means --all-versions has been enabled
		if isinstance(version, tuple):
			for ver in version:
				print_search(pkg, ver, pkg_list_check)
		else:
			print_search(pkg, version, pkg_list_check)

	return bool(pkg_list_check)


def print_search(pkg: Package, version: Version, pkg_list_check: list[Package]) -> None:
	"""Print the search results to the terminal."""
	pkg_list_check.append(pkg)
	print(
		ascii_replace(
			set_search_description(
				set_search_installed(
					set_search_origin(
						f"{color(pkg.name, 'GREEN')} {color(version.version, 'BLUE')}",
						version,
					),
					pkg,
					version,
				),
				version,
			)
		),
		end="\n\n",
	)


def set_search_origin(line: str, version: Version) -> str:
	"""Return the provided string with the origin information."""
	if origin := version._cand.file_list[0][0]:
		if origin.component == "now":
			return _("{package} [local]").format(package=line)
		return f"{line} [{origin.label}/{origin.codename} {origin.component}]"
	return line


def set_search_installed(line: str, pkg: Package, version: Version) -> str:
	"""Return the provided string with install and upgrade information."""
	if version.is_installed and pkg.is_upgradable:
		# NOTE: Formatting looks as below:
		# NOTE: vim 2:8.2.3995-1+b2 [Debian/sid main]
		# NOTE: ├── is installed and upgradable to 2:8.2.4659-1
		# NOTE: └── Vi IMproved - enhanced vi editor
		return f"{line}\n{TOP_LINE} " + _(
			"is installed and upgradable to {version}"
		).format(
			version=color(pkg_candidate(pkg).version, "BLUE"),
		)
	if version == pkg.candidate and pkg.is_upgradable:
		# NOTE: vim 2:8.2.4659-1 [Debian/sid main]
		# NOTE: ├── is upgradable from 2:8.2.3995-1+b2
		# NOTE: └── Vi IMproved - enhanced vi editor
		return f"{line}\n{TOP_LINE} " + _("is upgradable from {version}").format(
			version=color(pkg_installed(pkg).version, "BLUE")
		)
	if version.is_installed:
		# NOTE: vim 2:8.2.3995-1+b2 [Debian/sid main]
		# NOTE: ├── is installed
		# NOTE: └── Vi IMproved - enhanced vi editor
		return f"{line}\n{TOP_LINE} " + _("is installed")
	return line


def set_search_description(line: str, version: Version) -> str:
	"""Return the provided string with the package description."""
	records = version._translated_records
	if arguments.full and records:
		desc = "\n    ".join(records.long_desc.splitlines())
		return f"{line}\n{BOT_LINE} {desc}"
	if records:
		return f"{line}\n{BOT_LINE} {records.short_desc}"
	# NOTE: vim 2:8.2.3995-1+b2 [Debian/sid main]
	# NOTE: ├── is installed
	# NOTE: └── No Description
	no_desc = _("No Description")
	return f"{line}\n{BOT_LINE}{COLOR_CODES['ITALIC']} {no_desc}{COLOR_CODES['RESET']}"



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Functions related to the `show` command."""
from __future__ import annotations

import re
from pathlib import Path
from random import shuffle
from typing import List, cast

from apt.package import BaseDependency, Dependency, Package, Version

from nala import _, color
from nala.cache import Cache
from nala.constants import ERROR_PREFIX, NOTICE_PREFIX, PACSTALL_METADATA
from nala.debfile import NalaBaseDep
from nala.options import arguments
from nala.rich import ascii_replace
from nala.utils import term, unit_str

# NOTE: This is for `nala show`
# NOTE: header info looks like "Package: steam:i386"
SHOW_INFO = _("{header} {info}") + "\n"
URL_PATTERN = re.compile(r"(https?://.*?/.*?/)")


def show_main(num: int, pkg: Package) -> int:
	"""Orchestrate show command with support for all_versions."""
	if num:
		print(f"\n{'='*term.columns}\n")
	count = len(pkg.versions)
	versions = pkg.versions if arguments.all_versions else [pkg.candidate]
	for ver_num, ver in enumerate(versions):
		if ver is None:
			print(
				_("{pkg_name} has no candidate").format(
					pkg_name=color(pkg.name, "YELLOW")
				)
			)
			continue
		if ver_num and not num:
			print(f"\n{'='*term.columns}\n")
		count -= 1
		show_pkg(ver)
	return count


def show_pkg(candidate: Version) -> None:
	"""Start show functions."""
	pkg = candidate.package
	msg = f"{show_format(pkg, candidate)}\n{show_related(candidate)}"
	if candidate.homepage:
		msg += SHOW_INFO.format(header=color(_("Homepage:")), info=candidate.homepage)
	if candidate.size:
		msg += SHOW_INFO.format(
			header=color(_("Download-Size:")), info=unit_str(candidate.size)
		)
	msg += SHOW_INFO.format(
		header=color(_("APT-Sources:")), info=format_sources(candidate, pkg)
	)
	if candidate._translated_records:
		msg += SHOW_INFO.format(
			header=color(_("Description:")),
			info=ascii_replace(candidate._translated_records.long_desc),
		)
	print(msg.strip())


def show_related(candidate: Version) -> str:
	"""Show relational packages."""
	msg = ""
	if candidate.provides:
		msg += SHOW_INFO.format(
			header=color(_("Provides:")),
			info=show_dep([color(name, "GREEN") for name in candidate.provides]),
		)

	if candidate.enhances:
		msg += SHOW_INFO.format(
			header=color(_("Enhances:")),
			info=show_dep([color(pkg[0].name, "GREEN") for pkg in candidate.enhances]),
		)

	if candidate.dependencies:
		depends, pre_depends = split_deps(candidate.dependencies)
		if pre_depends:
			msg += SHOW_INFO.format(
				header=color(_("Pre-Depends:")), info=show_dep(pre_depends)
			)
		if depends:
			msg += SHOW_INFO.format(header=color(_("Depends:")), info=show_dep(depends))

	if candidate.recommends:
		msg += SHOW_INFO.format(
			header=color(_("Recommends:")), info=show_dep(candidate.recommends)
		)

	if candidate.suggests:
		msg += SHOW_INFO.format(
			header=color(_("Suggests:")), info=show_dep(candidate.suggests)
		)

	return msg + additional_related(candidate)


def additional_related(candidate: Version) -> str:
	"""Show breaks, conflicts, replaces."""
	msg = ""
	if replaces := candidate.get_dependencies("Replaces"):
		msg += SHOW_INFO.format(
			header=color(_("Replaces:")),
			info=show_dep([color(pkg[0].name, "GREEN") for pkg in replaces]),
		)
	if conflicts := candidate.get_dependencies("Conflicts"):
		msg += SHOW_INFO.format(header=color(_("Conflicts:")), info=show_dep(conflicts))
	if breaks := candidate.get_dependencies("Breaks"):
		msg += SHOW_INFO.format(header=color(_("Breaks:")), info=show_dep(breaks))
	return msg


def show_format(pkg: Package, candidate: Version) -> str:
	"""Format main section for show command."""
	installed = _("yes") if candidate.is_installed else _("no")
	essential = _("yes") if pkg.essential else _("no")
	maintainer = format_maintainer(f"{candidate.record.get('Maintainer')}".split())

	show = (
		SHOW_INFO.format(header=color(_("Package:")), info=color(pkg.name, "GREEN"))
		+ SHOW_INFO.format(
			header=color(_("Version:")), info=color(candidate.version, "BLUE")
		)
		+ SHOW_INFO.format(
			header=color(_("Architecture:")), info=candidate.architecture
		)
		+ SHOW_INFO.format(header=color(_("Installed:")), info=installed)
		+ SHOW_INFO.format(header=color(_("Priority:")), info=candidate.priority)
		+ SHOW_INFO.format(header=color(_("Essential:")), info=essential)
		+ SHOW_INFO.format(header=color(_("Section:")), info=candidate.section)
		+ SHOW_INFO.format(header=color(_("Source:")), info=candidate.source_name)
	)
	if origin := candidate.origins[0].origin:
		show += SHOW_INFO.format(header=color(_("Origin:")), info=origin)
	show += SHOW_INFO.format(header=color(_("Maintainer:")), info=maintainer)

	if original_maintainer := candidate.record.get("Original-Maintainer"):
		show += SHOW_INFO.format(
			header=color(_("Original-Maintainer:")),
			info=format_maintainer(f"{original_maintainer}".split()),
		)
	if bugs := candidate.record.get("Bugs"):
		show += SHOW_INFO.format(header=color(_("Bugs:")), info=bugs)
	if installed_size := candidate.installed_size:
		show += SHOW_INFO.format(
			header=color(_("Installed-Size:")), info=unit_str(installed_size)
		)
	return show.strip()


def format_maintainer(maintainer: list[str]) -> str:
	"""Format email in maintainer line."""
	maint_list = []
	for line in maintainer:
		if ">" in line:
			line = color(line[1:-1], "BLUE")
			line = color("<") + line + color(">")
		maint_list.append(line)
	return " ".join(maint_list)


def show_dep(dependency: list[Dependency] | list[str]) -> str:
	"""Print dependencies for show."""
	if isinstance(dependency[0], str):
		dependency.sort()
		if len(dependency) > 4:
			return "\n  " + "\n  ".join(cast(list[str], dependency))
		return ", ".join(cast(List[str], dependency))

	dependency = dedupe_deps(cast(List[Dependency], dependency))
	join_list = []
	msg = ""
	same_line = True
	if len(dependency) > 4:
		same_line = False
		msg += "\n"

	for dep_list in dependency:
		dep_print = ""
		for num, dep in enumerate(dep_list):
			assert isinstance(dep, BaseDependency)
			if num == 0:
				dep_print = format_dep(dep, num)
			else:
				dep_print += format_dep(dep, num)

		if same_line:
			join_list.append(dep_print.strip())
			continue
		msg += f"{dep_print}\n"

	return ", ".join(join_list) if same_line else msg.rstrip()


def format_dep(dep: BaseDependency | NalaBaseDep, iteration: int = 0) -> str:
	"""Format dependencies for show."""
	open_paren = color("(")
	close_paren = color(")")
	name = color(dep.name, "GREEN")
	if dep.rawtype in ("Breaks", "Conflicts"):
		name = color(dep.name, "RED")
	relation = color(dep.relation)
	version = color(dep.version, "BLUE")
	indent = color(" | ") if iteration > 0 else "  "

	final = f"{name} {open_paren}{relation} {version}{close_paren}"

	return indent + final if dep.relation else indent + name


def format_sources(candidate: Version, pkg: Package) -> str:
	"""Show apt sources."""
	origin = candidate.origins[0]
	if origin.archive == "now":
		return f"{get_local_source(pkg.shortname)}"
	return (
		f"{source_url(candidate.uris)} {origin.archive}/"
		f"{origin.component} {candidate.architecture} Packages"
	)


def source_url(uris: list[str]) -> str:
	"""Return the source url."""
	shuffle(uris)
	for mirror in uris:
		if regex := re.search(URL_PATTERN, mirror):
			return regex.group()
	return ""


def get_local_source(pkg_name: str) -> str:
	"""Determine the local source and return it."""
	postfixes = ("", "-deb", "-git", "-bin", "-app")
	for postfix in postfixes:
		metadata = PACSTALL_METADATA / (pkg_name + postfix)
		if metadata.exists():
			return parse_pacstall(metadata)
	return _("local install")


def parse_pacstall(pacdata: Path) -> str:
	"""Parse pacstall metadata file."""
	remote = "_remoterepo="
	# _remoterepo="https://github.com/pacstall/pacstall-programs"
	for line in pacdata.read_text().splitlines():
		if line.startswith(remote):
			index = line.index("=") + 1
			return color(line[index:].strip('"'), "BLUE")
	return color("https://github.com/pacstall/pacstall-programs", "BLUE")


def split_deps(
	depend_list: list[Dependency],
) -> tuple[list[Dependency], list[Dependency]]:
	"""Split dependencies into pre-depends and depends."""
	depends: list[Dependency] = []
	pre_depends: list[Dependency] = []
	for depend in depend_list:
		if depend[0].pre_depend:
			pre_depends.append(depend)
			continue
		depends.append(depend)
	return depends, pre_depends


def dedupe_deps(duplicates: list[Dependency]) -> list[Dependency]:
	"""Remove duplicate entries from a list while maintaining the order."""
	deduped = []
	dep_names = []
	for dep in duplicates:
		if dep.rawstr not in dep_names:
			dep_names.append(dep.rawstr)
			deduped.append(dep)
	return deduped


def additional_notice(additional_records: int) -> None:
	"""Print notice of additional records."""
	print(
		_(
			"{notice} There are {num} additional records. Please use the {switch} switch to see them."
		).format(
			notice=NOTICE_PREFIX,
			num=color(additional_records, "YELLOW"),
			switch=color("'-a'", "YELLOW"),
		)
	)


def pkg_not_found(pkg_name: str, cache: Cache, not_found: list[str]) -> None:
	"""Add not found errors to the list."""
	if cache.is_virtual_package(pkg_name):
		return
	if cache.is_secret_virtual(pkg_name):
		not_found.append(
			_("{error} {package} has no version to show").format(
				error=ERROR_PREFIX, package=pkg_name
			)
		)
		return
	not_found.append(
		_("{error} {package} not found").format(error=ERROR_PREFIX, package=pkg_name)
	)
